//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Device/Detector/FlatDetector.cpp
//! @brief     Implements class FlatDetector.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Device/Detector/FlatDetector.h"
#include "Base/Axis/Frame.h"
#include "Base/Axis/FrameTrafo.h"
#include "Base/Axis/MakeScale.h"
#include "Base/Axis/Scale.h"
#include "Base/Const/Units.h"
#include "Base/Pixel/RectangularPixel.h"
#include "Base/Util/Assert.h"
#include "Device/Beam/Beam.h"
#include "Device/Resolution/IDetectorResolution.h"
#include <numbers>
using std::numbers::pi;
#include <iostream>

FlatDetector::FlatDetector(size_t nxbins, size_t nybins, double width, double height,
                           const Beam& beam, NominalNormal nominalNormal, double distance,
                           double offcenter_w, double offcenter_h, double pitch, double yaw,
                           double roll)
    : m_nominalNormal(nominalNormal)
    , m_u0(width / 2 + offcenter_w)
    , m_v0(height / 2 + offcenter_h)
    , m_distance(distance)
{
    ASSERT(m_distance > 0);
    ASSERT(!pitch); // TODO: implement
    ASSERT(!yaw);   // TODO: implement
    ASSERT(!roll);  // TODO: implement

    if (m_nominalNormal == X)
        m_normal_to_detector = R3(m_distance, 0.0, 0.0);

    else if (m_nominalNormal == T)
        m_normal_to_detector = m_distance * beam.ki().unit_or_throw();

    else if (m_nominalNormal == R)
        m_normal_to_detector = m_distance * beam.k_reflected().unit_or_throw();

    else
        ASSERT(false);

    const R3 ym_unit(0.0, -1.0, 0.0);
    const double d2 = m_normal_to_detector.dot(m_normal_to_detector);
    const R3 u_direction = d2 * ym_unit - ym_unit.dot(m_normal_to_detector) * m_normal_to_detector;
    m_u_unit = u_direction.unit_or_throw();
    m_v_unit = m_u_unit.cross(m_normal_to_detector).unit_or_throw();

    // for use in lambda captures:
    const R3 D = m_normal_to_detector;
    const double u0 = m_u0;
    const double v0 = m_v0;
    const R3 u_unit = m_u_unit;
    const R3 v_unit = m_v_unit;
    const R3 ki = beam.ki();
    const double k0 = beam.wavenumber();

    auto pos_in_det = [=](double u, double v) -> R3 {
        return D + (u - u0) * u_unit + (v - v0) * v_unit;
    };

    trafo_map_t x_trafos = {
        {"q_y (1/nm)",
         [=](double u) -> double {
             const R3 p = pos_in_det(u, 0);
             const R3 kf = k0 * p.unit_or_throw();
             const R3 q = kf - ki;
             return -q.y();
         }},
        {"u (mm)", [](double u) -> double { return u; }},
        {"nxbins", [=](double u) -> double { return nxbins / width * u; }},
        {"phi_f (deg)",
         [=](double u) -> double {
             const R3 p = pos_in_det(u, 0);
             return atan2(-p.y(), p.x()) * (180 / pi);
         }},
    };
    trafo_map_t y_trafos = {
        {"q_z (1/nm)",
         [=](double v) -> double {
             const R3 p = pos_in_det(0, v);
             const R3 kf = k0 * p.unit_or_throw();
             const R3 q = kf - ki;
             return q.z();
         }},
        {"v (mm)", [](double v) -> double { return v; }},
        {"nybins", [=](double v) -> double { return nybins / height * v; }},
        {"alpha_f (deg)",
         [=](double v) -> double {
             const R3 p = pos_in_det(0, v);
             return atan2(p.z(), p.x()) * (180 / pi);
         }},
    };
    Scale* x = newEquiDivision("u (mm)", nxbins, 0.0, width);
    Scale* y = newEquiDivision("v (mm)", nybins, 0.0, height);
    x->setTrafos(x_trafos);
    y->setTrafos(y_trafos);
    Frame* f = new Frame(x, y);
    f->addTrafo(new FrameTrafo{"q (1/nm)",
                               {{"q_y (1/nm)"}, {"q_z (1/nm)"}},
                               {[=](double u) -> double {
                                    const R3 p = pos_in_det(u, 0);
                                    const R3 kf = k0 * p.unit_or_throw();
                                    const R3 q = kf - ki;
                                    return -q.y();
                                },
                                [=](double v) -> double {
                                    const R3 p = pos_in_det(0, v);
                                    const R3 kf = k0 * p.unit_or_throw();
                                    const R3 q = kf - ki;
                                    return q.z();
                                }},
                               true});
    f->addTrafo(
        new FrameTrafo{"u,v (mm)",
                       {{"u (mm)"}, {"v (mm)"}},
                       {[](double u) -> double { return u; }, [](double v) -> double { return v; }},
                       false});
    f->addTrafo(new FrameTrafo{"nbins",
                               {{"nubins"}, {"nvbins"}},
                               {[=](double u) -> double { return nxbins / width * u; },
                                [=](double v) -> double { return nybins / height * v; }},
                               false});
    f->addTrafo(new FrameTrafo{"angles (deg)",
                               {{"phi_f (deg)"}, {"alpha_f (deg)"}},
                               {[=](double u) -> double {
                                    const R3 p = pos_in_det(u, 0);
                                    return atan2(-p.y(), p.x()) * (180 / pi);
                                },
                                [=](double v) -> double {
                                    const R3 p = pos_in_det(0, v);
                                    return atan2(p.z(), p.x()) * (180 / pi);
                                }},
                               true});
    setFrame(f);
}

FlatDetector::FlatDetector(const FlatDetector& other) = default;

FlatDetector::~FlatDetector() = default;

FlatDetector* FlatDetector::clone() const
{
    return new FlatDetector(*this);
}

double FlatDetector::width() const
{
    return axis(0).span();
}

double FlatDetector::height() const
{
    return axis(1).span();
}

size_t FlatDetector::xSize() const
{
    return axis(0).size();
}

size_t FlatDetector::ySize() const
{
    return axis(1).size();
}

double FlatDetector::getU0() const
{
    return m_u0;
}

double FlatDetector::getV0() const
{
    return m_v0;
}

double FlatDetector::getDistance() const
{
    return m_distance;
}

FlatDetector::NominalNormal FlatDetector::getDetectorArrangment() const
{
    return m_nominalNormal;
}

const RectangularPixel* FlatDetector::regionOfInterestPixel() const
{
    const auto uAxisBoundsROI = regionOfInterestBounds(0);
    const auto vAxisBoundsROI = regionOfInterestBounds(1);

    const double u_min = uAxisBoundsROI.first;
    const double v_min = vAxisBoundsROI.first;
    const double width = uAxisBoundsROI.second - uAxisBoundsROI.first;
    const double height = vAxisBoundsROI.second - vAxisBoundsROI.first;

    const R3 corner_position(m_normal_to_detector + (u_min - m_u0) * m_u_unit
                             + (v_min - m_v0) * m_v_unit);
    const R3 uaxis_vector = width * m_u_unit;
    const R3 vaxis_vector = height * m_v_unit;
    return new RectangularPixel(corner_position, uaxis_vector, vaxis_vector);
}

const IPixel* FlatDetector::createPixel(size_t index) const
{
    const Scale& u_axis = axis(0);
    const Scale& v_axis = axis(1);
    const size_t u_index = axisBinIndex(index, 0);
    const size_t v_index = axisBinIndex(index, 1);

    const Bin1D u_bin = u_axis.bin(u_index);
    const Bin1D v_bin = v_axis.bin(v_index);
    const R3 corner_position(m_normal_to_detector + (u_bin.lowerBound() - m_u0) * m_u_unit
                             + (v_bin.lowerBound() - m_v0) * m_v_unit);
    const R3 width = u_bin.binSize() * m_u_unit;
    const R3 height = v_bin.binSize() * m_v_unit;
    return new RectangularPixel(corner_position, width, height);
}

// ISSUE #396. cover by test. currently insensitive to change in k_spec.
size_t FlatDetector::indexOfSpecular(const Beam& beam) const
{
    const R3 k_spec = beam.k_reflected();
    const R3 normal_unit = m_normal_to_detector.unit_or_throw();
    const double kd = k_spec.dot(normal_unit);
    if (kd <= 0.0)
        return totalSize();
    ASSERT(m_distance != 0);
    const R3 rpix = k_spec * (m_distance / kd);
    const double u = rpix.dot(m_u_unit) + m_u0;
    const double v = rpix.dot(m_v_unit) + m_v0;
    const Scale& u_axis = axis(0); // the x axis, GISAS only
    const Scale& v_axis = axis(1); // the y axis, in reflectometer plane
    if (!u_axis.rangeComprises(u) || !v_axis.rangeComprises(v))
        return totalSize();
    return getGlobalIndex(u_axis.closestIndex(u), v_axis.closestIndex(v));
}

std::string FlatDetector::normalName() const
{
    switch (m_nominalNormal) {
    case X:
        return "X";
    case T:
        return "T";
    case R:
        return "R";
    default:
        ASSERT(false);
    }
}
